#!/usr/bin/env python
#
# api: python
# type: application
# title: streamtuner2
# description: directory browser for internet radio / audio streams
# depends: gtk, pygtk, xml.dom.minidom, threading
# version: 1.9.6
# author: mario
# license: public domain
# url: http://freshmeat.net/p/streamtuner2
#
#
# Streamtuner2 is a GUI browser for internet radio directories. Various
# providers can be added, and streaming stations are usually grouped into
# music genres or categories.
#
# It's an independent rewrite of streamtuner1 in a scripting language. So
# it can be more easily extended and fixed. It still lacks many features
# and some might get left out even.
# Therefore streamtuner2 should be considered just an alternative to the
# original. It's not necessarily a replacement for everyone.
#
# As it is in a very early stage (just two weeks of development), it also
# isn't much more stable. It works rudimentarily, the GUI is very wacky
# however, and it has neither been optimized for memory use or performance
# yet. The internal workings are supposed to develop further... (not best
# structure right now.)  And documentation is lacking.
#
#
#  features:
#   - less functionality than original C streamtuner 0.99.x
#   - treeview lists are created from datamap[] structure and stream{} dict
#   - channel categories are built-in defaults (can be freshened up however)
#   - config vars and cache data is stored as JSON in ~/.config/streamtuner2
#
#  missing:
#   - favicons
#   - right click context menus
#   - oh, and the main menu (at least it is incomplete now)
#   - config panel
#   - localization
#   - standard channel class, instead of shoutcast as base
#   - per-channel .play() method instead of using url hints
#   - search feature
#   - bookmarks
#   - 
#
#  security notes:
#   - limited format checking for shoutcast parsing (just a fragile regex)
#   - no XML/SGML entity decoding, urls might be left encoded
#   - little integrity checking for .pls files
#   - use of eval() for pseudojson.read() - only local files are read however
#   - verification of downloaded favicon data files (ico/png) isn't performed
#   - no proper parsing of XML for xiph.org
#
#  help needed on:
#   - any of the above
#   - logo
#   - more professional python app structure
#   - packaging (deb/rpm)
#



import sys, os
import pygtk, gtk, gtk.glade
import urllib
#import cPickle
import re
#import psyco
#psyco.full()
from threading import Thread
import xml.dom.minidom


# this represents the main window
# and also contains most application behaviour
class StreamTunerTwo(gtk.glade.XML):

        # constructor
        def __init__(self):

            # instantiate gtk/glade widgets in current object
            gtk.glade.XML.__init__(self, sys.path[0]+"/st2.glade")
                # gtk.glade.xml_new_from_buffer...
      
            # bind gtk/glade event names to functions
            self.signal_autoconnect({
                "gtk_main_quit" : self.gtk_main_quit,
                # treeviews / notebook
                "on_stream_row_activated" : self.on_shoutcast_list_row_activated,
                "on_category_clicked": self.on_category_clicked,
                "on_notebook_channels_switch_page": self.channel_switch,
                # toolbar
                "on_play_clicked" : self.on_play_clicked,
                "on_record_clicked": self.on_record_clicked,
                "on_homepage_stream_clicked": self.on_homepage_stream_clicked,
                "on_reload_clicked": self.on_reload_clicked,
                "on_stop_clicked": self.on_stop_clicked,
                "on_homepage_channel_clicked" : self.on_homepage_channel_clicked,
                # menu
                "menu_toolbar_standard": lambda(w): (self.toolbar.unset_style(), self.toolbar.unset_icon_size()),
                "menu_toolbar_style_icons": lambda(w): (self.toolbar.set_style(gtk.TOOLBAR_ICONS)),
                "menu_toolbar_style_both": lambda(w): (self.toolbar.set_style(gtk.TOOLBAR_BOTH)),
                "menu_toolbar_size_small": lambda(w): (self.toolbar.set_icon_size(gtk.ICON_SIZE_SMALL_TOOLBAR)),
                "menu_toolbar_size_medium": lambda(w): (self.toolbar.set_icon_size(gtk.ICON_SIZE_DND)),
                "menu_toolbar_size_large": lambda(w): (self.toolbar.set_icon_size(gtk.ICON_SIZE_DIALOG)),
                # else
                "app_state": self.app_state,
                "bookmark": self.bookmark,
            })

            # channel modules
            self.c = {}  # needs to exist before instantiating modules, else __getattr__ access on self.c[] creates race condition
            self.c = {
              "bookmarks": bookmarks(parent=self),
              "default": None,
              "shoutcast": shoutcast(parent=self),
              "google": None,
              "live365": None,
              "xiph": xiph(parent=self),
              "search": None,
            }
            self.cn = ["bookmarks", "default", "shoutcast", "google", "live365", "xiph", "search"]    # order of channel notebook tabs
            self.cc = "shoutcast"    # currently selected channel name (as index in self.c{})
            

            # set some defaults
            winlayout = conf.get("window")
            if (winlayout):
                mygtk.app_restore(self, winlayout)
                0
            #
            # actually display main window
            self.win_streamtuner2.show()
            self.working = []
          

        # shortcut fo glade.get_widget(),
        # allows access to widgets as direct attributes instead of using .get_widget()
        # or looking in self.c[] for the named channel plugins
        def __getattr__(self, name):
            if (self.c.has_key(name)):
                return self.c[name]
            else:
                return self.get_widget(name)
            # or gives an error, if neither exists



        # play button
        def on_play_clicked(self, widget):
            url = self.selected(self.c[self.cc].gtk_list, "url");
            #format = self.selected(self.c[self.cc].gtk_list, "format");
            format = "audio/mp3"
            action.play(url, format)

        # streamripper
        def on_record_clicked(self, widget):
            url = self.selected(self.c[self.cc].gtk_list, "url")
            action.record(url, "audio/mp3")

        # browse stream
        def on_homepage_stream_clicked(self, widget):
            url = self.selected(self.c[self.cc].gtk_list, "homepage")             
            action.browser(url)
             
        # browse channel
        def on_homepage_channel_clicked(self, widget):
            action.browser(self.c[self.cc].base_url)
            
        # reload stream list in current channel-category
        def on_reload_clicked(self, widget):
            thread = Thread(self.c[self.cc].reload())
            thread.start()
            self.working.append(thread)
        # stop
        def on_stop_clicked(self, widget):
            thread = self.working.pop()
            thread.stop()

        # doubleclick in stream list
        def on_shoutcast_list_row_activated(self, widget, event, *more):
            self.on_play_clicked(widget)
        
        # click in category list
        def on_category_clicked(self, widget, event, *more):
            (model, iter) = widget.get_selection().get_selected()
            category = model.get_value(iter, 0)
            self.c[self.cc].load(category)
            pass

        # fetches varname from treeview selection            
        def selected(self, widget, name):
            (model, iter) = widget.get_selection().get_selected()
            if (model and iter):
                liststore_index = self.c[self.cc].rowmap.index(name)
                return model.get_value(iter, liststore_index)
            else:
                pass
                
        # shortcut to statusbar
        def status(self, text=None):
            # prepare statusbar            
            if (not self.__dict__.has_key("sbar_cid")):
                self.sbar_cid = self.statusbar.get_context_id("messages")
                self.sbar_msg = []
            # set or remove text
            if (not text):
                self.statusbar.pop(self.sbar_cid)
            else:
                #self.sbar_msg.append(
                self.statusbar.push(self.sbar_cid, text)
            pass
            
        def gtk_main_quit(self, widget):
            gtk.main_quit()

        def app_state(self, widget):
            conf.save("window", mygtk.app_state(wTree=self, widgetnames=[
               "win_streamtuner2","shoutcast_list","shoutcast_cat","toolbar","notebook_channels"
            ]))

        def channel_switch(self, notebook, page, page_num, *args):
            self.cc = self.cn[page_num]
            
        def bookmark(self, widget):
            (model, iter) = self.c[self.cc].gtk_list.get_selection().get_selected()
            row = {}
            for i,id in enumerate(self.c[self.cc].rowmap):
                val = model.get_value(iter, i)
                row[id] = val
            self.bookmarks.add(row)







# SHOUTcast data module
class shoutcast:

        # desc
        api = "st2"
        module = "shoutcast"
        version = 0.7
        base_url = "http://classic.shoutcast.com/"

        # content
        categories = ["default", [], 'TopTen', [], 'Alternative', ['College', 'Emo', 'Hardcore', 'Industrial', 'Punk', 'Ska'], 'Americana', ['Bluegrass', 'Blues', 'Cajun', 'Folk'], 'Classical', ['Contemporary', 'Opera', 'Symphonic'], 'Country', ['Bluegrass', 'New Country', 'Western Swing'], 'Electronic', ['Acid Jazz', 'Ambient', 'Breakbeat', 'Downtempo', 'Drum and Bass', 'House', 'Trance', 'Techno'], 'Hip Hop', ['Alternative', 'Hardcore', 'New School', 'Old School', 'Turntablism'], 'Jazz', ['Acid Jazz', 'Big Band', 'Classic', 'Latin', 'Smooth', 'Swing'], 'Pop/Rock', ['70s', '80s', 'Classic', 'Metal', 'Oldies', 'Pop', 'Rock', 'Top 40'], 'R&B/Soul', ['Classic', 'Contemporary', 'Funk', 'Smooth', 'Urban'], 'Spiritual', ['Alternative', 'Country', 'Gospel', 'Pop', 'Rock'], 'Spoken', ['Comedy', 'Spoken Word', 'Talk'], 'World', ['African', 'Asian', 'European', 'Latin', 'Middle Eastern', 'Reggae'], 'Other/Mixed', ['Eclectic', 'Film', 'Instrumental']]
        current = "default"
        streams = {}      #meta information dicts
        liststore = {}    #gtk data structure
        gtk_list = None   #gtk widget
        gtk_cat = None    #gtk widget

        # mapping of stream{} data into gtk treeview/treestore representation
        datamap = [
           # coltitle   width	[ datasrc key, type, renderer, attrs ]	[cellrenderer2], ...
           ["",		20,	["state",	str,  "pixbuf",	{}],	],
           ["genre",	65,	['genre',	str,	"t",	{}],	],
           ["title",	275,	["title",	str,    "text",	{"strikethrough":11}],    ],#["favicon",gtk.gdk.Pixbuf,"pixbuf",{}], ],
           ["playing",	185,	["playing",	str,	"text",	{"strikethrough":11}],	],
           ["listeners", 45,	["listeners",	int,	"t",	{"strikethrough":11}],	],
           ["max",	45,	["max",		int,	"t",	{}],	],
           ["bitrate",	35,	["bitrate",	int,	"t",	{}],	],
           ["homepage",	160,	["homepage",	str,	"t",	{"underline":10}],	],
           ["stream",	25,	["url",		str,	"t",	{"strikethrough":11}],	],
           [False,	0,	["format",	str,	None,	{}],	],
           [False,	0,	["favourite",	bool,	None,	{}],	],
           [False,	0,	["deleted",	bool,	None,	{}],	],
        ]        
        rowmap = []   # [state,genre,title,...] field enumeration still needed separately
        

        # constructor
        def __init__(self, parent=None):

            # only if streamtuner2 is run in graphical mode        
            if (parent):
                self.cache()
                self.gui(parent)
            pass
        

        # read previous channel/stream data, if there is any
        def cache(self):
            # stream list
            cache = conf.get("cache/" + self.module)
            if (cache):
                self.streams = cache
            # categories
            cache = conf.get("cache/categories_" + self.module)
            if (cache):
                self.categories = cache
            pass

        # switch stream category,
        # load data,
        # update treeview content
        def load(self, category, force=False):
            # get data from cache or download
            if (force or not self.streams.has_key(category)):
                new_streams = self.update_streams(category)
                if (len(new_streams)):
                    # don't lose forgotten streams
                    self.streams[category] = new_streams + self.deleted_streams(new_streams, self.streams.get(category,[]))
                    # save in cache
                    conf.save("cache/" + self.module, self.streams)
                    # invalidate gtk list cache
                    if (self.liststore.has_key(category)):
                        del self.liststore[category]
                else:
                    # parse error
                    self.parent.status("category parsed empty.")
                    self.streams[category] = [{"title":"no contents found on directory server","bitrate":0,"max":0,"listeners":0,"playing":"error","favourite":0,"deleted":0}]
                    print "oooops, parser returned nothing for category " + category
                    pass
            # assign to treeview model
            self.streams["default"] = []
            if (self.liststore.has_key(category)):
                self.gtk_list.set_model(self.liststore[category])
            else:
                self.liststore[category] = \
                mygtk.columns(self.gtk_list, self.datamap, self.display(self.streams[category]))
            self.current = category
            pass

            
        # finds differences in new/old streamlist, marks deleted with flag
        def deleted_streams(self, new, old):
            diff = []
            new = [row.get("url","http://example.com/") for row in new]
            for row in old:
                if (row.has_key("url") and (row["url"] not in new)):
                    row["deleted"] = 1
                    diff.append(row)
            return diff
        
        # update data for display
        def display(self, streams):
            for i,row in enumerate(streams):
                                            # oh my, at least it's working
                                            # first the bookmarks module isn't fully registered at instatiantion in parent.c{} - might want to do that step by step rather
                                            # then display() is called too early to take effect - load() & co should actually be postponed to when a notebook tab gets selected first
                if (conf.show_bookmarks and self.parent.c.has_key("bookmarks") and self.parent.c["bookmarks"].is_in(streams[i]["url"])):
                    streams[i]["favourite"] = 1
                if (row.get("favourite")):
                    streams[i]["state"] = gtk.STOCK_ABOUT
                if (row.get("deleted")):
                    streams[i]["state"] = gtk.STOCK_DELETE
            return streams
            

        # reload current stream from web directory
        def reload(self):
            self.load(self.current, force=1)
            

        # initialize Gtk widgets / data objects
        def gui(self, parent):

            # save reference to main window/glade API
            self.parent = parent
            self.gtk_list = parent.get_widget(self.module+"_list")
            self.gtk_cat = parent.get_widget(self.module+"_cat")
            
            # category tree
            mygtk.tree(self.gtk_cat, self.categories, title="category", icon=gtk.STOCK_OPEN);
            
            # prepare stream list
            if (not self.rowmap):
                for row in self.datamap:
                    for x in xrange(2, len(row)):
                        self.rowmap.append(row[x][0])
            self.load(self.current)

            
        # extracts the category list from shoutcast.com,
        # the text/html format always being:
        #        <OPTION VALUE="Classical">Classical
        #                <OPTION VALUE="Contemporary"> - Contemporary
        #                <OPTION VALUE="Opera"> - Opera
        # sub-categories are indented by two tabs, maincats just by one
        def update_categories(self):
            f = urllib.urlopen(self.base_url + "") 
            html = f.read(1<<17)
            self.categories = ["default"]
            sub = []
            for uu in re.findall(r'<OPTION VALUE="(.+?)">( - )?(.+)', html):
                name = urllib.unquote(uu[0])
                # main category
                if (len(uu[1]) == 0):
                    self.categories.append(name)
                    sub = []
                    self.categories.append(sub)
                # sub entry
                else:
                    sub.append(name)
            # it's done
            #print self.categories
            conf.save("cache/categories_shoutcast", self.categories)
            pass



        # downloads stream list from shoutcast for given category
        def update_streams(self, cat, search=""):
        
            # regular patterns
            rx_stream = re.compile("""
                 href="/(sbin/shoutcast-playlist.pls[^"]+)">.+?
                 size="1"><b>\[([^\]]+?)\]</.+?
                 <a.id="listlinks".target="_scurl".href="([^"]+)">([^<>]+)</a>.+?
                 >Now.Playing:</font>\s*([^<>]+)</font>.+?
                 (\d+)/(\d+).+?
                 >(\d+)</font>.+?
                 <!--.color.AAC\+.Type.data.-->.+?>\s*([\w\d]+)[+]*<
                 """, re.S|re.I|re.X)
            rx_next = re.compile(">\[ Next \]</a>")
            
            # loop
            entries = []
            next = 0
            while (next<2000):
            
               # get page
               url = self.base_url + "directory/index.phtml?sgenre="+urllib.quote(cat)+"&maxbitrate=999&numresult=100&startat="+str(next)+"&s="+search
               #print url
               html = http.get(url)
               #print "rx"              
               
               # extract entries
               self.parent.status("parsing document...")
               for uu in rx_stream.findall(html):
                   #print uu
                   entries += [{
                       "url": self.base_url + uu[0],
                       "genre": uu[1],
                       "homepage": http.fix_url(uu[2]),
                       "title": uu[3],
                       "playing": uu[4],
                       "listeners": int(uu[5]),
                       "max": int(uu[6]),
                       "bitrate": int(uu[7]),
                       "format": "audio/" + uu[8].lower(),
                   }]
               self.parent.status()
               
               # more pages to load?
               if (re.search(rx_next, html)):
                  next += 100
               else:
                  next = 99999
            
            #fin
            return entries
            
          
# I wonder what that is for  
class xiph (shoutcast):

        # desc
        api = "st2"
        module = "xiph"
        version = 0.1
        base_url = "http://dir.xiph.org/"
        yp = "yp.xml"

        # content
        categories = ["all", [],           ]
        current = "all"


        # downloads stream list from shoutcast for given category
        def update_categories(self):
            pass


        def x(self, entry, name):
            e = entry.getElementsByTagName(name)
            if (e):
                if (e[0].childNodes):
                    return e[0].childNodes[0].data

        # downloads stream list from shoutcast for given category
        def update_streams(self, cat, search=""):
        
            #-- get data
            yp = http.get(self.base_url + self.yp, 1<<22)
            
            #-- extract
            l = []
            for entry in xml.dom.minidom.parseString(yp).getElementsByTagName("entry"):
                l.append({
                    "title": str(self.x(entry, "server_name")),
                    "url": str(self.x(entry, "listen_url")),
                    "format": str(self.x(entry, "server_type")),
                    "bitrate": 0, #self.x(entry, "bitrate"),
                    "channels": str(self.x(entry, "channels")),
                    "samplerate": str(self.x(entry, "samplerate")),
                    "genre": str(self.x(entry, "genre")),
                    "playing": str(self.x(entry, "current_song")),
                    "listeners": 0,
                    "max": 0,
                    "homepage": "",
                })
                
            # filter by category:
            if (cat != "all"):
                0
            
            #done   
            return l            


#-- favourite lists
# this module lists static content from ~/.config/streamtuner2/bookmarks.json only,
# its data list is queries by other plugins to add 'star' icons
class bookmarks(shoutcast):            
        # desc
        api = "st2"
        module = "bookmarks"
        version = 0.1
        base_url = "file:.config/streamtuner2/bookmarks.json"
        # content
        categories = ["favourite",]
        current = "favourite"
        streams = {"favourite":[]}
        urls = []

        def update_categories(self):
            pass
        def update_streams(self):
            pass
        def cache(self):
            # stream list
            cache = conf.get(self.module)  # becomes "bookmarks.json"
            if (cache):
                self.streams = cache
            pass
        def is_in(self, url, once=1):
            if (not self.urls):
                self.urls = [row["url"] for row in self.streams["favourite"]]
            return url in self.urls
        def add(self, row):
            self.streams["favourite"].append(row)
            row["favourite"] = 1  # send status back into source streams list
            self.urls.append(row["url"])
            self.save()
            pass
        def save(self):
            conf.save(self.module, self.streams)
            pass
        def load(self, category, force=False):
            self.liststore[category] = \
            mygtk.columns(self.gtk_list, self.datamap, self.display(self.streams[category]))
            pass
            




#-- global configuration data               ---------------------------------------------
class Config:

        # start
        def __init__(self):

            # prepare
            self.xdg()
            self.defaults()
            
            # settings from last session
            last = self.get("settings")
            if (last):
                for key,value in last.iteritems():
                    self.__dict__[key] = value
            #
            self.save("settings")


        # some defaults
        def defaults(self):
            self.browser = "sensible-browser"
            self.play = {
               "audio/mp3": "audacious %m3u",	# %u for url to .pls, %g for downloaded .m3u
               "audio/aac": "vlc %u",
               "audio/x-pn-realaudio": "vlc %u",
               "*/*": "vlc %u",
            }
            self.record = {
               "*/*": "x-terminal-emulator -c streamripper %g",
            }
            self.tmp = os.environ.get("TEMP", "/tmp")
            
            self.show_bookmarks = 1
        
            
        # http://standards.freedesktop.org/basedir-spec/basedir-spec-0.6.html
        def xdg(self):
            home = os.environ.get("HOME", "/tmp")
            config = os.environ.get("XDG_CONFIG_HOME", ".config")
            
            # storage dir
            self.dir = home + "/" + config + "/" + "streamtuner2"
            
            # create if necessary
            if (not os.path.exists(self.dir)):
                os.mkdir(self.dir)
           

        # store some configuration list/dict into a file                
        def save(self, name="settings", data=None):
            name = name + ".json"
            if (data == None):
                data = self.__dict__
            # check for subdir
            if (name.find("/") > 0):
                subdir = name[0:name.find("/")]
                subdir = self.dir + "/" + subdir
                if (not os.path.exists(subdir)):
                    os.mkdir(subdir)
            # write                        
            file = self.dir + "/" + name
            #cPickle.dump(data, open(file, "w"))
            pseudojson.save(file, data)
            pass


        # retrieve data from config file            
        def get(self, name):
            name = name + ".json"
            file = self.dir + "/" + name
            if (os.path.exists(file)):
                # return cPickle.load(open(file, "r"))
                return pseudojson.read(file)
            pass
        

   


#-- media actions                           ---------------------------------------------
#
# implements "play" and "record" methods,
# but also "browser" for web URLs
#        
class action:

        # calls player for stream url and format
        @staticmethod
        def play(url, format="audio/mp3"):
            cmd = conf.play.get(format, conf.play.get("*/*", "vlc %u"))
            print "play", url, cmd
            if (cmd):
                cmd = action.interpol(cmd, url)
                print cmd
                os.system(cmd + " &")

        # streamripper
        @staticmethod
        def record(url, format="audio/mp3"):
            cmd = conf.record.get(format, conf.record.get("*/*", None))
            if (cmd):
                cmd = action.interpol(cmd, url)
                os.system(cmd + " &")

        # streamripper
        @staticmethod
        def browser(url):
            cmd = conf.record.get("browser", "firefox")
            os.system(cmd + " '" + url + "' &")
            
        # replaces instances of %u, %pls, %g, %m3u with urls or local filenames
        @staticmethod
        def interpol(cmd, url):
            if (re.search("%u|%url|%pls", cmd)):
                cmd = re.sub("%u|%url|%pls", "'"+url+"'", cmd)
            if (re.search("%g|%m3u", cmd)):
                cmd = re.sub("%g|%m3u", action.m3u(url), cmd)
            return cmd
            
        # download a .pls resource and extract urls
        @staticmethod
        def pls(url):
            text = http.get(url)
            """
            [playlist]
            numberofentries=1
            File1=http://85.17.174.181:9006
            Title1=(#1 - 138/200) Pop Fusion Radio. Your Hit Music Channel
            Length1=-1
            Version=2
            """
            return re.findall("\s*File\d*\s*=\s*(\w+://[^\s]+)", text, re.I)
            # currently misses out on the titles            
            
        # create a local .m3u file from it
        @staticmethod
        def m3u(pls, channelname="shoutcast"):
            # extract stream address from .pls URL
            if (re.search("\.pls", pls)):       #audio/x-scpls
                url_list = action.pls(pls)
            elif (re.search("\.m3u", pls)):	#audio/x-mpegurl
                url_list = re.findall("(http://[^\s]+)", pls, re.I)
            else:  # just assume it was a direct mp3/ogg streamserver link
                url_list = [ http.fix_url(pls) ]

            print url_list            
            # use shoutcast unique stream id if available
            stream_id = re.search("(\d+)", pls)
            if (stream_id):
                stream_id = stream_id.group(1)
            else:  # makes little sense in this context to use random value (?)
                stream_id = "XXXXXX"
            channelname = main.cc

            # output URL list to temporary .m3u file
            if (len(url_list)):
                tmp_fn = conf.tmp+"/streamtuner2."+channelname+"."+stream_id+".m3u"
                f = open(tmp_fn, "w")
                f.write("#M3U\n")
                for url in url_list:
                   f.write(http.fix_url(url) + "\n")
                f.close()
                # return path/name of temporary file
                return tmp_fn
            else:
                print "error, there were no URLs in ", pls
                pass
            
            
            



# simplified gtk constructors               ---------------------------------------------
class mygtk:
             
        #-- fill a treeview
        #
        # Adds treeviewcolumns/cellrenderers and liststore from a data dictionary.
        # It can be called in two steps: first for the datamap, and later together
        # with the data. If new data gets applied, the columns aren't recreated.
        #
        # The columns are created according to the datamap, which describes cell
        # mapping and layout.
        #
        # datamap = [  #  title   width     dictkey     type,  renderer,  attrs  
        #               ["name",   150,  ["titlerow",   str,    "text",    {} ]  ],
        #               [False,     0,   ["interndat",  int,     None,     {} ]  ],
        #               ["desc",   200,  ["descriptn",  str,    "text",    {} ],  ["icon",str,"pixbuf",{}]  ],
        #
        # Practically you only need widget, datamap and entries. The rowmap parameter
        # was hacked in to export the keylist, if it's required elsewhere (clicked
        # lookups). pix_entry= can be given, if entries contains image-filenames where
        # gdk.Pixmaps are required - then the display-code automatically converts them.
        #
        @staticmethod
        def columns(widget, datamap=[], entries=[], pix_entry=False):

            # create treeviewcolumns?
            if (not widget.get_column(0)):
                # loop through titles
                datapos = 0
                for n_col,desc in enumerate(datamap):
                                    
                    # check for title
                    if (type(desc[0]) != str):
                        datapos += 1  # if there is none, this is just an undisplayed data column
                        continue
                    # new tvcolumn
                    col = gtk.TreeViewColumn(desc[0])  # title
                    col.set_resizable(True)
                    # width
                    if (desc[1] > 0):
                        col.set_sizing(gtk.TREE_VIEW_COLUMN_FIXED)
                        col.set_fixed_width(desc[1])

                    # loop through cells
                    for var in xrange(2, len(desc)):
                        cell = desc[var]
                        # cell renderer
                        if (cell[2] == "pixbuf"):
                            rend = gtk.CellRendererPixbuf()  # img cell
                            if (cell[1] == str):
                                cell[3]["stock_id"] = datapos  # for stock icons
                                expand = False
                            else:
                                pix_entry = datapos
                                cell[3]["pixbuf"] = datapos
                        else:
                            rend = gtk.CellRendererText()    # text cell
                            cell[3]["text"] = datapos
                            col.set_sort_column_id(datapos)  # only on textual cells
   
                        # attach cell to column
                        col.pack_end(rend, expand=cell[3].get("expand",True))
                        # apply attributes
                        for attr,val in cell[3].iteritems():
                            col.add_attribute(rend, attr, val)
                        # next
                        datapos += 1

                    # add column to treeview
                    widget.append_column(col)
                # finalize widget
                widget.set_search_column(2)   #??
                widget.set_reorderable(True)
               
            # add data?
            if (entries):
                #- expand datamap            
                vartypes = []  #(str, str, bool, str, int, int, gtk.gdk.Pixbuf, str, int)
                rowmap = []    #["title", "desc", "bookmarked", "name", "count", "max", "img", ...]
                if (not rowmap):
                    for desc in datamap:
                        for var in xrange(2, len(desc)):
                            vartypes += [desc[var][1]]  # content types
                            rowmap += [desc[var][0]]    # dict{} column keys in entries[] list
                # create gtk array storage
                ls = gtk.ListStore(*vartypes)   # could be TreeStore, too

                # generate ordered list from dictionary, using rowmap association
                typedefaults = {str:"", bool:False, int:0, gtk.gdk.Pixbuf:gtk.gdk.pixbuf_new_from_data("\077\0\0\077",gtk.gdk.COLORSPACE_RGB,True,8,1,1,4)}
                if gtk.gdk.Pixbuf in vartypes:
                    pix_entry = vartypes.index(gtk.gdk.Pixbuf) 
                for row in entries:
                    # sort data into array
                    row = [row.get(skey,typedefaults[vartypes[i]]) for i,skey in enumerate(rowmap)]
                    # autotransform string -> gtk image object
                    if (pix_entry and type(row[pix_entry]) == str):
                        row[pix_entry] = gtk.gdk.pixbuf_new_from_file(row[pix_entry])
                    # add
                    ls.append(row)   # had to be adapted for real TreeStore (would require additional input)
                # apply array to widget
                widget.set_model(ls)
                return ls
                
            pass


        # simple two-level treeview display in one column
        # with entries = [main,[sub,sub], title,[...],...]
        @staticmethod     
        def tree(widget, entries, title="category", icon=gtk.STOCK_DIRECTORY):
            # list types
            ls = gtk.TreeStore(str, str)
            # add entries
            for entry in entries:
                if (type(entry) == str):
                    main = ls.append(None, [entry, icon])
                else:
                    for sub_title in entry:
                        ls.append(main, [sub_title, icon])
            # just one column
            tvcolumn = gtk.TreeViewColumn(title);
            widget.append_column(tvcolumn)
            # inner display: icon & string
            pix = gtk.CellRendererPixbuf()
            txt = gtk.CellRendererText()
            # position
            tvcolumn.pack_start(pix, expand=False)
            tvcolumn.pack_end(txt, expand=True)
            # select array content source in treestore
            tvcolumn.set_attributes(pix, stock_id=1)
            tvcolumn.set_attributes(txt, text=0)
            # finalize
            widget.set_model(ls)
            tvcolumn.set_sort_column_id(0)
            #tvcolumn.set_sizing(gtk.TREE_VIEW_COLUMN_FIXED)
            #tvcolumn.set_fixed_width(125])
            widget.set_search_column(0)
            widget.expand_all()
            #widget.expand_row("3", False)
            #print widget.row_expanded("3")
            return ls


        #-- save window size and widget properties
        # needs a list of widgetnames
        # e.g. pickle.dump(mygtk.app_state(...), open(os.environ["HOME"]+"/.config/app_winstate", "w"))
        @staticmethod
        def app_state(wTree, widgetnames=["window1", "treeview2", "vbox17"]):
            r = {} # restore array
            for wn in widgetnames:
                r[wn] = {}
                w = wTree.get_widget(wn)
                t = type(w)
                #print wn, w, t
                # extract different information from individual widget types
                if t == gtk.Window:
                    r[wn]["size"] = list(w.get_size())
                if t == gtk.Widget:
                    r[wn]["name"] = w.get_name()
                # gtk.TreeView
                if t == gtk.TreeView:
                    r[wn]["columns:width"] = []
                    for col in w.get_columns():
                        r[wn]["columns:width"].append( col.get_width() )
                    # - Rows
                    r[wn]["rows:expanded"] = []
                    for i in xrange(0,50):
                        if w.row_expanded(str(i)):
                            r[wn]["rows:expanded"].append(i)
                # gtk.Toolbar
                if t == gtk.Toolbar:
                    r[wn]["icon_size"] = int(w.get_icon_size())
                    r[wn]["style"] = int(w.get_style())
                # gtk.Notebook
                if t == gtk.Notebook:
                    r[wn]["page"] = w.get_current_page()
            print r
            return r

        #-- restore window and widget properties
        # requires only the previously saved widget state dict
        @staticmethod
        def app_restore(wTree, r=None):
            for wn in r.keys():  # widgetnames
                w = wTree.get_widget(wn)
                t = type(w)
                for method,args in r[wn].iteritems():
                    # gtk.Window
                    if method == "size":
                        w.resize(args[0], args[1])
                    # gtk.TreeView
                    if method == "columns:width":
                        for i,col in enumerate(w.get_columns()):
                            col.set_fixed_width(args[i])
                    #  - Rows
                    if method == "rows:expanded":
                        w.collapse_all()
                        for i in args:
                            w.expand_row(str(i), False)                        
                    # gtk.Toolbar
                    if method == "icon_size":
                        w.set_icon_size(args)
                    if method == "style":
                        w.set_style(args)
                    # gtk.Notebook
                    if method == "page":
                        w.set_current_page(args)

            pass





#-- writes/reads pseudo-JSON                ---------------------------------------------
# the basic python data types dict,list,str,int are mostly identical to json,
# therefore a basic str() conversion is enough for writing,
# for reading the more bothersome eval() is used
# - it's however no severe security problem here, because we're just reading
#   local config files (written by us) and accept no data from outside / web
class pseudojson:
        @staticmethod
        def save(fn, obj):
            f = open(fn, "w")
            f.write(str(obj).replace("'}, ", "'\n},\n ").replace("', ", "',\n "))  # .replace slows down and might corrupt the data ( unwanted whitespace)
            f.close()      #.replace("': [{'", "':\n[\n{'")
            pass
        @staticmethod
        def read(fn):
            f = open(fn, "r")
            obj = eval(f.read(1<<24))   # max 16 MB
            f.close()
            return obj
            
            

#-- url download                            ---------------------------------------------
class http:

        #-- GET
        @staticmethod
        def get(url, maxsize=1<<19):
            #main.status("downloading " + url)
            f = urllib.urlopen(url)
            content = f.read(maxsize)
            #main.status()
            return content

        #-- fix invalid URLs
        @staticmethod
        def fix_url(url):
            if len(url):
                if (url.find("://") < 0):
                    url = "http://" + url
                if (url.find("/", 10) < 0):
                    url = url + "/"
            return url
            
   

#-- run main                                ---------------------------------------------
if __name__ == "__main__":
    conf = Config()
    st2 = main = StreamTunerTwo()
    gtk.main()

